﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Media.Imaging;
using System.Windows.Threading;
using TPLSamples.Infrastructure;
using TPLSamples.Utility;

namespace TPLSamples.Models
{
	public class Sampler : DomainObject
	{
		private readonly Lazy<ICollection<Sample>> lazySamples = new Lazy<ICollection<Sample>>(() => new List<Sample>());
		public ICollection<Sample> Samples
		{
			get { return lazySamples.Value; }
		}

		string imagePath = "";
		public string ImagePath
		{
			get { return this.imagePath; }
			set
			{
				if (this.imagePath != value)
				{
					this.ValidateProperty("ImagePath", value);
					this.imagePath = value;
					this.RaisePropertyChanged("ImagePath");
					SetImageSources();
				}
			}
		}

		BitmapSource sourceImage;
		public BitmapSource SourceImage
		{
			get { return sourceImage; }
			set
			{
				if (this.sourceImage != value)
				{
					this.ValidateProperty("SourceImage", value);
					this.sourceImage = value;
					this.RaisePropertyChanged("SourceImage");
				}
			}
		}

		BitmapSource destinationImage;
		public BitmapSource DestinationImage
		{
			get { return destinationImage; }
			set
			{
				if (this.destinationImage != value)
				{
					this.ValidateProperty("DestinationImage", value);
					this.destinationImage = value;
					this.RaisePropertyChanged("DestinationImage");
				}
			}
		}

		string results = "";
		public string Results
		{
			get { return this.results; }
			set
			{
				if (this.results != value)
				{
					this.ValidateProperty("Results", value);
					this.results = value;
					this.RaisePropertyChanged("Results");
				}
			}
		}

		public bool IsComplete
		{
			get
			{
				return !string.IsNullOrEmpty(this.ImagePath);
			}
		}

		public Sampler(params Sample[] samples)
		{
			if (File.Exists("../../../Edison.jpg"))
			{
				ImagePath = new FileInfo("../../../Edison.jpg").FullName;
			}

			foreach (var sample in samples)
			{
				Samples.Add(sample);
			}
		}

		protected override void ValidateProperty(string propertyName, object value)
		{
			if (propertyName == "ImagePath")
			{
				var errors = new List<string>();
				string path = value as string;
				if (string.IsNullOrEmpty(path))
				{
					errors.Add("Image path must be set.");
				}
				else if (!File.Exists(path))
				{
					errors.Add("The image indicated does not exist.");
				}

				this.ErrorsContainer.SetErrors(propertyName, errors);
			}
			else
			{
				base.ValidateProperty(propertyName, value);
			}
		}


		public async Task RunSamples()
		{
			await Task.Run(() =>
			{
				System.Drawing.Bitmap bmp = null;
				foreach (var sample in Samples)
				{
					if(!sample.IsEnabled)
						continue;

					if (sample.ImageRequired)
					{
						UpdateResults("Reseting Image");
						ResetDestinationImage();
						bmp = GetBitmapFromDestinationImage();
					}
					UpdateResults("Starting " + sample.SampleName);
					sample.Run(bmp, UpdateResults);
					UpdateResults("Completed " + sample.SampleName);
					UpdateResults(sample.SampleName + " ran in " + sample.RunTime.ToString() + Environment.NewLine);
				}

				if (bmp != null)
				{
					SetDestinationImageFromBitmap(bmp);
				}
			});
		}
		
		private void UpdateResults(string result)
		{
			Results += result + Environment.NewLine;
		}
		

		private void SetImageSources()
		{
			if(!File.Exists(ImagePath))
				return;

			BitmapImage sourceBitmapImage = new BitmapImage();
			sourceBitmapImage.BeginInit();
			sourceBitmapImage.UriSource = new Uri(ImagePath);
			sourceBitmapImage.EndInit();
			SourceImage = sourceBitmapImage;

			ResetDestinationImage();
		}

		public void ResetDestinationImage()
		{
			if(!File.Exists(ImagePath))
				return;

			if (Application.Current.Dispatcher.Thread != Thread.CurrentThread)
			{ 
				Application.Current.Dispatcher.Invoke(() => 
				{
					ResetDestinationImage();
				});
				return;
			}
			BitmapImage destBitmapImage = new BitmapImage();
			destBitmapImage.BeginInit();
			destBitmapImage.UriSource = new Uri(ImagePath);
			destBitmapImage.EndInit();
			DestinationImage = destBitmapImage;
		}

		private System.Drawing.Bitmap GetBitmapFromDestinationImage()
		{
			System.Drawing.Bitmap bmp = null;
			if (Application.Current.Dispatcher.Thread != Thread.CurrentThread)
			{
				Application.Current.Dispatcher.Invoke(() => 
				{
					bmp = GetBitmapFromDestinationImage();
				});
				return bmp;
			}
			bmp = new System.Drawing.Bitmap(DestinationImage.PixelWidth, DestinationImage.PixelHeight, System.Drawing.Imaging.PixelFormat.Format32bppPArgb);
			System.Drawing.Imaging.BitmapData data = bmp.LockBits(
			  new System.Drawing.Rectangle(System.Drawing.Point.Empty, bmp.Size),
			  System.Drawing.Imaging.ImageLockMode.WriteOnly,
			  System.Drawing.Imaging.PixelFormat.Format32bppPArgb);
			DestinationImage.CopyPixels(
			  System.Windows.Int32Rect.Empty,
			  data.Scan0,
			  data.Height * data.Stride,
			  data.Stride);
			bmp.UnlockBits(data);

			return bmp;
		}


		private void SetDestinationImageFromBitmap(System.Drawing.Bitmap bitmap)
		{
			if (bitmap == null)
				throw new ArgumentNullException("bitmap");
			
			if (Application.Current.Dispatcher.Thread != Thread.CurrentThread)
			{
				Application.Current.Dispatcher.Invoke(() => 
				{
					SetDestinationImageFromBitmap(bitmap);
				});
				return;
			}

			IntPtr hBitmap = bitmap.GetHbitmap();

			try
			{
				DestinationImage = System.Windows.Interop.Imaging.CreateBitmapSourceFromHBitmap(
					hBitmap,
					IntPtr.Zero,
					Int32Rect.Empty,
					BitmapSizeOptions.FromEmptyOptions());
			}
			finally
			{
                NativeMethods.DeleteObject(hBitmap);
			}
		}
	}
}
